﻿///<reference path="scripts/babylon.max.js" />

/*
 * The file for part two of Babylon.js's Bowling tutorial for MSDN magazine.
 * Created (with great pleasure) by Raanan Weber, https://blog.raananweber.com
 * Follow me on twitter https://twitter.com/RaananW
 * Or drop me a line - info@raananweber.com
 */

//Physics vars
var bowlingBallWeight = 6.8;
var bowlingPinWeight = 1.5;

//camera collision detection

function enableCameraCollision(camera, scene) {
    //Enable gravity on the scene. Should be similar to earth's gravity.
    scene.gravity = new BABYLON.Vector3(0, -0.98, 0);

    // Enable collisions globally
    scene.collisionsEnabled = true;

    //Enable collision detection and gravity on the free camera
    camera.checkCollisions = true;
    camera.applyGravity = true;

    //Set the player size, the camera's ellipsoid.
    camera.ellipsoid = new BABYLON.Vector3(0.3, 0.8, 0.3);
}

function enableMeshesCollision(meshes) {
    meshes.forEach(function (mesh) {
        mesh.checkCollisions = true;
    });
}

function enablePhysics(scene, floor, lane, gutters, ball, pins) {

    //Set the physics state for all static objects
    scene.enablePhysics(new BABYLON.Vector3(0, -9.8, 0), new BABYLON.OimoJSPlugin());
    floor.setPhysicsState(BABYLON.PhysicsEngine.BoxImpostor, { mass: 0, friction: 0, restitution: 0 });
    lane.setPhysicsState(BABYLON.PhysicsEngine.BoxImpostor, { mass: 0, friction: 0, restitution: 0 });
    gutters.forEach(function (gutter) {
        gutter.setPhysicsState(BABYLON.PhysicsEngine.BoxImpostor, { mass: 0, friction: 0, restitution: 0 });
    })
    //Init the pins
    initPins(scene, pins);
    //Update the ball's position and its physics state
    var ballPosition = scene.activeCameras[0].position.clone();
    ballPosition.y /= 2;
    resetBall(ball, ballPosition);
}

function initFollowCamera(scene, ball) {
    //Add the active camera to the active cameras array
    scene.activeCameras.push(scene.activeCamera);

    //Create the follow object, which will always be in the ball's position
    var followObject = BABYLON.Mesh.CreateBox("followObject", 0.001, scene);
    followObject.position = ball.position;

    //Create the follow camera
    var followCamera = new BABYLON.FollowCamera("followCamera", ball.position, scene);
    followCamera.target = followObject;
    followCamera.viewport = new BABYLON.Viewport(0, 0, 0.3, 0.3);
    followCamera.radius = 1.5; // how far from the object to follow
    followCamera.heightOffset = 1; // how high above the object to place the camera
    followCamera.rotationOffset = 180; // the viewing angle
    followCamera.cameraAcceleration = 0.5 // how fast to move
    followCamera.maxCameraSpeed = 20;

    return followCamera;
}

/*
 * This function should be used only if the 3 needed mp3 files were added to the project.
 * I couldn't find public domain bowling audio effects and therefore couldn't include them in the project.
 */
function initAudio(ball, pins, scene) {
    //init the ambient sound
    var atmosphere = new BABYLON.Sound("Ambient", "./Assets/ambient.mp3", scene, null, {
        loop: true,
        autoplay: true
    });
    //lower its volume
    atmosphere.setVolume(0.2);

    //init the rolling sound
    var rollingSound = new BABYLON.Sound("rolling", "./Assets/rolling.mp3", scene, null, {
        loop: true,
        autoplay: false
    });
    //attach it to the ball
    rollingSound.attachToMesh(ball);

    //init the hit sound (will be attached to the first pin later
    var hitSound = new BABYLON.Sound("hit", "./Assets/hit.mp3", scene);

    return {
        atmosphere: atmosphere,
        rollingSound: rollingSound,
        hitSound: hitSound
    }
}

var strengthCounter = 8;

function configureThrowAction(scene, ball, pins, floor, followCamera, sounds) {

    //the function that increases the strength counter to max 26
    var counterUp = function () {
        strengthCounter = Math.min(strengthCounter + .5, 26);
    }

    scene.onPointerUp = function (evt, pickInfo) {
        //only left click
        if (!(evt.button == 0)) return;
        //stop increasing the strength counter
        scene.unregisterBeforeRender(counterUp);

        if (pickInfo.pickedMesh == floor) return;

        //calculate throw direction
        var direction = pickInfo.pickedPoint.subtract(ball.position).normalize();

        //Apply the strength counter, with a max value of 25.
        var impulse = direction.scale(strengthCounter);

        ball.applyImpulse(impulse, ball.getAbsolutePosition());
        //start the rolling sound
        sounds.rollingSound && sounds.rollingSound.play();

        //Add the follow camera
        scene.activeCameras.push(followCamera);
        //set the height at 1 meter
        followCamera.heightOffset = 1.0;

        scene.registerBeforeRender(function ballCheck() {
            //"flying" effect of the follow camera (will lower its height slowly);
            followCamera.heightOffset = Math.max(followCamera.heightOffset - 0.01, 0.2);

            if (ball.intersectsMesh(floor, false)) {
                //the ball intersects with the floor, stop checking its position. 
                scene.unregisterBeforeRender(ballCheck);
                //stop the rolling sound
                sounds.rollingSound && sounds.rollingSound.stop();

                //Let the ball roll around for 1.5 seconds before resetting it.
                setTimeout(function () {
                    //remove the follow camera from the active cameras array
                    scene.activeCameras.splice(scene.activeCameras.indexOf(followCamera), 1);

                    //Reset the ball in a new position
                    var newPosition = scene.activeCameras[0].position.clone();
                    newPosition.y /= 2;
                    resetBall(ball, newPosition);

                }, 1500);
            }
        });

        //attach the audio to the 1st pin's position
        sounds.hitSound && sounds.hitSound.attachToMesh(pins[0]);

        //Check intersections with the pins
        scene.registerBeforeRender(function ballIntersectsPins() {
            var intersects = pins.some(function (pin) {
                return ball.intersectsMesh(pin, false);
            });
            if (intersects) {
                //The ball touched at least one of the pins.
                scene.unregisterBeforeRender(ballIntersectsPins);
                //play the hit audio
                sounds.hitSound && sounds.hitSound.play();
            }
        });

        strengthCounter = 8;
    }

    scene.onPointerDown = function (evt, pickInfo) {
        //only left click
        if (!(evt.button == 0)) return;
        scene.registerBeforeRender(counterUp);
    }
}

function resetBall(ball, newPosition) {
    ball.position.copyFrom(newPosition);
    //Set the ball's physics state with the sphere impostor
    ball.setPhysicsState(BABYLON.PhysicsEngine.SphereImpostor, { mass: bowlingBallWeight, friction: 0, restitution: 0 });
}

function updateScoreboard(scene, pins) {

    var score = 10;

    //Create the dynamic texture
    var dyntex = new BABYLON.DynamicTexture("dt", 512, scene);
    //Create the plane on which the texture will be projected
    var scoreboard = BABYLON.Mesh.CreatePlane("scoreboard", 5, scene, false);
    //set its position
    scoreboard.position.z = 40;
    //Set its material with the dynamic texture as diffuse texture
    scoreboard.material = new BABYLON.StandardMaterial("scoradboardMat", scene);
    scoreboard.material.diffuseTexture = dyntex;
    //Say hello :-)
    clear();

    scene.registerBeforeRender(function () {
        var newScore = 10 - checkPins(pins);
        //If pins down value was updated, repaint the texture
        if (newScore != score) {
            score = newScore;
            dyntex.clear();
            dyntex.drawText(score + " pins down", 40, 100, "bold 72px Arial", "white", "black");
        }
        //if strength counter is > 8, print the strength
        if (strengthCounter > 8) {
            dyntex.clear();
            dyntex.drawText("strength:" + strengthCounter, 40, 100, "bold 72px Arial", "white", "black");
        }
    });

    function clear() {
        score = 10;
        initPins(scene, pins);
        dyntex.clear();
        dyntex.drawText("hello!", 140, 100, "bold 72px Arial", "white", "black");
    }

    //Pressing "r" will reset the scene. Lovely!
    scene.actionManager.registerAction(new BABYLON.ExecuteCodeAction({ trigger: BABYLON.ActionManager.OnKeyUpTrigger, parameter: "r" }, clear));
}

function initPins(scene, pins) {
    //initialize the pins
    pins.forEach(function (pin, idx) {
        //set their render group to be the same as the ball's
        pin.renderingGroupId = 1;
        //set their position. Clone - so the original position won't be updated.
        pin.position = pinPositions[idx].clone();
        //set their quaternion. This is needed to work with the physics engine.
        pin.rotationQuaternion = new BABYLON.Quaternion(0, 0, 0, 1);
        //raise them above the ground to create a "falling" effect like in a real bowling lane
        pin.position.y += 0.2;
        //set their physics state
        pin.body = pin.setPhysicsState(BABYLON.PhysicsEngine.CylinderImpostor, { mass: bowlingPinWeight, friction: 0, restitution: 0 });
    });
}

function checkPins(pins) {
    var pinsStanding = 0;
    pins.forEach(function (pin, idx) {
        //is the pin still standing on top of the lane?
        if (BABYLON.Tools.WithinEpsilon(pinYPosition, pin.position.y, 0.01)) {
            pinsStanding++;
        }
    });
    return pinsStanding;
}